/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/RAW.h,v 1.2 1999/01/31 22:33:52 benkovsk Exp benkovsk $
 */

/**********************************************************/
#ifndef __RAW_H
#define __RAW_H

#include <string.h>

#include <gtk/gtk.h>
#include <libgimp/gimp.h>

#ifdef RAW_WITH_GUILE
#include <guile/gh.h>
#endif

#define PLUGIN_LOAD_NAME   "file_raw_load"
#define PLUGIN_SAVE_NAME   "file_raw_save"
#define PLUGIN_LOAD_MENU   "<Load>/Raw"
#define PLUGIN_SAVE_MENU   "<Save>/Raw"

#define WARNING g_print
#define ERROR   g_error

#define MAX_INBUF_SIZE  1024
#define DEF_MAX_ARG_SIZE 1024  /* default for format->max_arg_size */

typedef gchar  VarName;        /* type of variable name */
typedef gint   VarNum;         /* type of numeric variable */
typedef gfloat VarFloat;       /* type of float variable */
typedef gchar  VarText;        /* type of string variable */

#ifdef RAW_WITH_GUILE
#define SCM2NUM  gh_scm2int
#define NUM2SCM  gh_int2scm
#endif

/**********************************************************/
/**********************************************************/

/*
 * type of expressions found in format spec files
 */

typedef enum {
  EXP_NUM = 0,          /* numeric - VarNum */
  EXP_FLOAT,            /* float - FIXME: not used now */
  EXP_STR,              /* string - VarText */
  EXP_VAR,              /* variable reference */
  EXP_LISP              /* LISP expression */
} ExpressionType;

/*
 * presentation type of subplugin argument ('ask' item)
 */

typedef enum {
  ASK_ENTRY = 0,        /* entry field */
  ASK_SCALE,            /* (numeric/float) scale slider field */
  ASK_TOGGLE            /* (numeric) check button */
} AskType;

/*
 * context for which given directive will be processed. It has effect
 *   only for 'ask', 'set' and 'hdr' directives just now
 */

typedef enum {
  CTX_BOTH = 0,         /* directive is valid during both save and load */
  CTX_SAVE,             /* directive is valid during save */
  CTX_LOAD              /* directive is valid during load */
} ContextType;

/*
 * Phase numbers. Phases serve for determining the right moment when
 *   the set item (var definition) should be evaluated
 * FIXME: add another phases
 */

typedef enum {
  PHASE_BEFHDR = 30     /* immediately  before call to (create|read)_header */
} PhaseType;

#define  PHASE_DEFAULT   PHASE_BEFHDR   /* use this phase if 0 was specified */

/* 
 * modes for (save|load)_pixel_x functions
 * FIXME: should be local in data.c
 */

typedef enum {          
  OP_NORMAL = 0,        /* xlat the pixels */
  OP_INIT,              /* prepare for xlatting pixels (calc invariants) */
  OP_DONE               /* done w/ xlatting pixels (flush buffers) */
} OperType;

/*
 * storage type of item in the file header. ('hdr' item)
 * FIXME: add another types
 */

typedef enum {
  HDR_LITTLEEND = 0,   /* little endian (least significant byte first) number */
  HDR_BIGEND,          /* big endian (most significant byte first) number */
  HDR_STRING           /* string */
} HdrItemType;


/**********************************************************/
/**********************************************************/

/*
 * struc to hold expression specified in format spec file
 */

typedef struct
{
  ExpressionType  type;     /* expression type */
  VarNum          num;      /* value for type EXP_NUM */
  VarFloat        fnum;     /* value for type EXP_FLOAT; FIXME: not used now */
  VarText         *text;    /* value if expression type is anything else */
  int             len;      /* length of text */
} Expression;

/*
 * struc containing format specific plugin parameters. These are requested
 * from user or fetched from PDB/subplugin args/plugin sp_args/...
 */

typedef struct
{
  VarName      *name;       /* var into which it will be stored */
  Expression   *label;      /* label to use in dialog box */
  Expression   *value;      /* current/default value */
  Expression   *help;       /* tooltip; FIXME: not used */
  AskType      type;        /* presentation type (for dialog box) */
  Expression   **constraints;  /* constraints to be applied during input */
  gint         constraint_cnt; /* # of constraints */
  ContextType  context;     /* when use this parameter */
} AskItem;

/*
 * struc describing piece of data in image file header
 */

typedef struct
{
  Expression      *offset;      /* offset from the header start */
  Expression      *size;        /* size of the item (bytes) */
  HdrItemType     type;         /* item type */
  Expression      *value;       /* item value */
  ContextType     context;      /* when use this header item */
} HdrItem;

/*
 * struc describing variable to be set in runtime by other means
 * than 'ask' or 'hdr' item
 */

typedef struct
{
  VarName       *name;          /* variable name */
  Expression    *value;         /* next variable value */
  PhaseType     phase;          /* phase in which it has to be evaluated */
  ContextType   context;        /* context in which to do the assignment */
  gboolean      processed;      /* this item has been processed already */
} SetItem;

/*
 * Raw Format - this struc describes user specified format of raw file
 *   this structure is created upon reading the format from file
 *   in configured directories
 */

typedef struct
{
  char        *file;            /* format file name */
  char        *id;              /* format ident for PDB */
  char        *name;            /* format name */
  char        *desc;            /* description for PDB */
  char        *author;          /* format author */
  char        *maintainer;      /* format maintainer */
  char        *date;            /* release date / version */
  char        *ext;             /* preferred file extension */
  char        *magic;           /* file magic */
  char        *loadmenu;        /* location in load menu tree */
  char        *savemenu;        /* location in save menu tree */
  guint32     allowed_types;    /* allowed image types FIXME: not used */
  char        *type;            /* ditto, for gimp_register_procedure */
  gint        max_arg_size;     /* buffer size for gimp_get_data() */
  Expression  *headersize;      /* size of img file header */
  GList       *hdr_list;        /* list of var assignments: hdr */
  GList       *ask_list;        /* list of var assignments: ask */
  GList       *set_list;        /* list of var assignments: set */
  gint        ask_cnt_save;     /* number of ask items during save */
  gint        ask_cnt_load;     /* number of ask items during load */
  char        *header_fn;       /* name of fn to load/save header. FIXME */
  char        *pixel_fn;        /* name of fn to load/save pixel */
  char        *drawable_fn;     /* name of fn to load/save drawable */
  char        *image_fn;        /* name of fn to load/save image */
} RawFormat;


/* function to load/save file header */
typedef gboolean (*HeaderFn) (FILE *tgt, RawFormat *format);

/* function to convert pixels between internal and external representation */
typedef   void (*PixelFn)(guchar *out_buf, guchar *in_buf, gint *olen, gint *ilen, OperType op);

/* function to load/save single drawable */
typedef   gint (*DrawableFn)(FILE *tgt, PixelFn pixel_fn, GDrawable *drawable);

/* function to load/save whole/part-of image */
typedef   gint (*ImageFn)(FILE *tgt, DrawableFn drawable_fn, PixelFn pixel_fn, gint32 image_ID, GDrawable *drawable);

/*
 * item in table describing available functions for saving/loading
 * file headers
 */

typedef struct
{
  char    *name;
  HeaderFn save_fn;
  HeaderFn load_fn;
} HeaderFnRec;

/*
 * item in table describing available functions for converting pixels
 * between internal (gimp) and external (raw file) representations
 */

typedef struct
{
  char     *name;                /* function name, as used in format file */
  PixelFn  save_fn;              /* function to save pixel */
  PixelFn  load_fn;              /* function to load pixel */
} PixelFnRec;

/*
 * item in table describing available functions for saving/loading
 * drawables
 */

typedef struct
{
  char        *name;             /* function name, as used in format file */
  DrawableFn  save_fn;           /* function to save drawable */
  DrawableFn  load_fn;           /* function to load drawable */
} DrawableFnRec;

/*
 * item in table describing available functions for saving/loading
 * whole images
 */

typedef struct
{
  char      *name;               /* function name, as used in format file */
  ImageFn   save_fn;             /* function to save image */
  ImageFn   load_fn;             /* function to load image */
} ImageFnRec;


/**********************************************************/
/**********************************************************/

/* FIXME: should be local */
extern char       *raw_home_path;
extern GList      *rawformat_list;
extern RawFormat  *selected_format;
extern GHashTable *variables;

extern VarNum     width;
extern VarNum     height;
extern int        alpha;
extern int        ncolors;
extern int        *indexno;
extern int        indexed;
extern int        bpp;
extern int        color;
extern int        alpha_threshold;
extern int        image_type;

/* FIXME: should be static */
extern PixelFnRec       pixel_fn_rec[];
extern gint             pixel_fn_rec_cnt;
extern DrawableFnRec    drawable_fn_rec[];
extern gint             drawable_fn_rec_cnt;
extern ImageFnRec       image_fn_rec[];
extern gint             image_fn_rec_cnt;


/**********************************************************/
/**********************************************************/

/* subplugin.c */
gchar*
subplugin_get_name (RawFormat *format, ContextType context);

gboolean
subplugin_get_args_from_pdb (RawFormat *format, ContextType context);

void
subplugin_set_args_to_pdb (RawFormat *format, ContextType context);

void
subplugin_install_all (void);

gboolean
subplugin_get_args (RawFormat *format, 
		    ContextType context, 
		    GParam *args, 
		    gint nargs);

gboolean
subplugin_get_args_from_string (RawFormat *format, 
				ContextType context, 
				gchar *argstr);

gchar*
subplugin_set_args_to_string (RawFormat *format, 
			      ContextType context);

void
subplugin_set_vars (RawFormat *format, ContextType context);


/* expression.c */
gchar*
string_escape (gchar *text, gint len);

gchar*
string_unescape (gchar *text, gint *len);

void
expression_free (Expression *value);

Expression*
expression_from_string (gchar *text, gchar **newtext);

gboolean
expression_assign_from_string (Expression *value, gchar *text);

gboolean
expression_assign_value_from_string (Expression *value, gchar *text);

gchar*
expression_to_string (Expression *value);

gchar*
expression_value_to_string (Expression *value);

ExpressionType
expression_evaluate (Expression *value, VarNum *num, VarText **text, gint *len);


/* FIXME: rename to expression_compare_assign () */

gboolean
expression_assign (Expression *lvalue, VarNum num, VarText *text, gint len);

Expression*
expression_copy (Expression *expr);

gint
expression_type_to_pdb (ExpressionType type);

#ifdef DEBUG
void
expression_print (char *name, Expression *value);
#endif

/* variables.c */
void
variables_init ();

void
variable_num_add (VarName *name, VarNum num);

void
variable_float_add (VarName *name, VarFloat fnum);

void
variable_text_add (VarName *name, VarText *text, gint len);

void
variable_expr_add (VarName *name, Expression *expr);

Expression*
variable_get (VarName *name);

gboolean
variable_get_num (VarName *name, VarNum *num);

gboolean
variable_get_float (VarName *name, VarFloat *fnum);

gboolean
variable_get_text (VarName *name, VarText **text, gint *len);


/* format.c */
RawFormat*
rawformat_find (char *id);

void
rawformat_load_all (void);

RawFormat*
rawformat_read (gchar *filename);

void
rawformat_free (RawFormat *format);

void
setitem_process_all (RawFormat *format, ContextType context, gint phase);


/* header.c */
void
header_find_function (ContextType context, gchar *header_name, 
		      HeaderFn *header_fn);


/* data.c */
void
data_find_function (ContextType context,
		    gchar *pixel_name, gchar *drawable_name, gchar *image_name,
		    PixelFn *pixel_fn, DrawableFn *drawable_fn, ImageFn *image_fn);

/**********************************************************/
#endif  /* ! __RAW_H */

/* End of file RAW.h */
