/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/format.c,v 1.2 1999/01/31 22:32:47 benkovsk Exp benkovsk $
 */

/**********************************************************/
#include <sys/stat.h>
#include <ctype.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "RAW.h"

#define RAW_FORMAT_MAGIC  "; RAW-Format Spec File v1.0"

/**********************************************************/
char       *raw_home_path = NULL;
GList      *rawformat_list = NULL;
RawFormat  *selected_format = NULL;

/**********************************************************/
static GList*
parse_formats_path (void);

static GList*
load_formats (GList *list);

/**********************************************************/
/**********************************************************/
void
askitem_free (AskItem *ask)
{
  gint   i;

  if (!ask) return;

  g_free (ask->name);
  expression_free (ask->label);
  expression_free (ask->help);

  for (i = ask->constraint_cnt; i > 0; i--)
    expression_free (ask->constraints[i-1]);

  if (ask->constraints) g_free (ask->constraints);

  g_free (ask);
}

/**********************************************************/
void
hdritem_free (HdrItem *hdr)
{
  if (hdr) {
    expression_free (hdr->offset);
    expression_free (hdr->size);
    expression_free (hdr->value);

    g_free (hdr);
  }
}

/**********************************************************/
void
setitem_free (SetItem *set)
{
  if (!set) return;

  g_free (set->name);
  expression_free (set->value);
  g_free (set);
}

/**********************************************************/
void
setitem_process_all (RawFormat *format, ContextType context, gint phase)
{
  GList   *slist;

  if (!format->set_list) return;

  slist = format->set_list->next;

  while (slist) {
    SetItem  *set = (SetItem*) slist->data;
    gchar    *text = NULL;
    gint     textlen = 0;
    VarNum    num = 0;

    slist = slist->next;

    if ((set->context == context || set->context == CTX_BOTH)
	&& (set->phase <= phase && ! set->processed)) {
      expression_evaluate (set->value, &num, &text, &textlen);

      if (text) variable_text_add (set->name, text, textlen);
      else variable_num_add (set->name, num);
    }
  }
}

/**********************************************************/
RawFormat*
rawformat_find (char *id)
{
  GList  *flist = rawformat_list->next;

  if (!id) return NULL;

  while (flist) {
    RawFormat  *format = (RawFormat*)flist->data;
    flist = flist->next;

    if (format->id && !strcmp (format->id, id)) return format;
  }

  return NULL;
}

/**********************************************************/
void
rawformat_load_all (void)
{
  GList  *flist;
  
  flist = parse_formats_path ();
  rawformat_list = load_formats (flist);
  g_list_free (flist);
}

/**********************************************************/
static RawFormat*
rawformat_new ()
{
  RawFormat  *format;

  format = g_new0 (RawFormat, 1);
  format->hdr_list = g_list_alloc ();
  format->ask_list = g_list_alloc ();
  format->set_list = g_list_alloc ();
  format->max_arg_size = DEF_MAX_ARG_SIZE;

  return format;
}

/**********************************************************/
static gboolean
rawformat_get_parsed_line (FILE *frm, gchar **name, Expression **n, 
			   gint *count, ContextType *context)
{
  gchar       inbuf[MAX_INBUF_SIZE];
  Expression  *expr, **np;
  gint        i;


  while (!feof (frm)) {
    gchar  *p = inbuf;

    inbuf[0]=0;
    fgets (inbuf, MAX_INBUF_SIZE, frm);

    p = inbuf + strlen (inbuf);
    if ((p > inbuf) && (*(p - 1) == '\n')) 
      *(p - 1) = '\0';

    p = inbuf;
    while (*p && *p<=' ') p++;
    if (!*p || *p == '#' || *p == ';') continue;


    p = strtok (p, ":");
    if (!p) {
      WARNING ("error\n");   /* FIXME */
      goto cleanup;
    }

    *context = CTX_BOTH;
    if (!strcasecmp (p, "save")) {
      *context = CTX_SAVE;
      p = strtok (NULL, ":");
    }
    else if (!strcasecmp (p, "load")) {
      *context = CTX_LOAD;
      p = strtok (NULL, ":");
    }

    if (!p) {
      WARNING ("error\n");   /* FIXME */
      goto cleanup;
    }

    
    *name = g_strdup (p);
    p += strlen (p) + 1;
    np = n;

    i = 0;
    while (*p && NULL != (expr = expression_from_string (p, &p))) {
      *np++ = expr;
      i++;
      if (*p) p++;   /* FIXME */
      else break;
    }

    *count = i;
    return TRUE;
  }

  *count = -1;
  return TRUE;

cleanup:
  *count = -1;
  return FALSE;
}

/**********************************************************/
/*
 * ... - tuples: [ expression type mask , ptr to target variable ]
 *       terminated w/ 0 instead of mask.
 * mask determines allowed EXP_* types for expr. values
 *   these are obtained as 1<<EXP_NUM | 1<<EXP_VAR, etc.
 * The mask could be or'ed with:
 *   0x10000 - copy whole expression, not only data (target is Expression**)
 *   0x20000 - replace variable name w/ constant numeric value
 */
static gboolean
set_param (gchar *name, Expression **n, gint count, ...)
{
  gboolean    rc = FALSE;
  Expression  *value;
  va_list     ap;
  gint        mask, i;
  gchar       **text;
  gint        *num;
  gfloat      *fnum;
  Expression  **expr;


  va_start (ap, count);

  i = 0;
  while (0 != (mask = va_arg (ap, gint))) {
    value = n[i];

    if (i >= count) {
      WARNING ("Too few arguments in '%s'\n", name);
      goto cleanup;
    }

    if (!((1 << value->type) & mask)) {
      WARNING ("Invalid argument type: %d !~ 0x%x\n", value->type, mask);
      goto cleanup;
    }

    if (mask & 0x10000) {
      expr = va_arg (ap, Expression**);
      if (*expr) WARNING ("%s: Duplicate definition ignored\n", name);
      else {
	*expr = value;
	n[i] = NULL;
      }
    }
    else {
      switch (value->type) {
      case EXP_STR:
	text = va_arg (ap, gchar**);
	if (*text) WARNING ("%s: Duplicate definition ignored\n", name);
	else {
	  *text = value->text;
	  value->text = NULL;
	}
	break;

      case EXP_NUM:
	num = va_arg (ap, VarNum*);
	*num = value->num;
	break;

      case EXP_FLOAT:
	fnum = va_arg (ap, VarFloat*);
	*fnum = value->fnum;
	break;

      case EXP_VAR:
	if (mask & 0x20000) {
	  num = va_arg (ap, gint*);
	  *num = 0;
	  if (!strcasecmp (value->text, "LITTLE")) *num = HDR_LITTLEEND;
	  else if (!strcasecmp (value->text, "L")) *num = HDR_LITTLEEND;
	  else if (!strcasecmp (value->text, "BIGEND")) *num = HDR_BIGEND;
	  else if (!strcasecmp (value->text, "B")) *num = HDR_BIGEND;
	  else if (!strcasecmp (value->text, "STRING")) *num = HDR_STRING;
	  else if (!strcasecmp (value->text, "S")) *num = HDR_STRING;
	  else if (!strcasecmp (value->text, "ENTRY")) *num = ASK_ENTRY;
	  else if (!strcasecmp (value->text, "TOGGLE")) *num = ASK_TOGGLE;
	  else if (!strcasecmp (value->text, "SCALE")) *num = ASK_SCALE;
	  else if (!strcasecmp (value->text, "PHASE_BEFHDR")) *num = PHASE_BEFHDR;
	  else if (!strcasecmp (value->text, "PHASE_DEFAULT")) *num = PHASE_DEFAULT;
	  else { 
	    WARNING ("Unrecognized constant: %s\n", value->text);
	    return FALSE;
	  }
	}
	else {
	  text = va_arg (ap, gchar**);
	  if (*text) WARNING ("%s: Duplicate definition ignored\n", name);
	  else {
	    *text = value->text;
	    value->text = NULL;
	  }
	}
	break;

      default:
	WARNING ("This expr type can't be used here\n");
	goto cleanup;
      }
    }

    i++;
  }

  rc = TRUE;

cleanup:
  va_end (ap);

  for (i=0; i<count; i++) 
    expression_free (n[i]);

  return rc;
}

/**********************************************************/
AskItem*
askitem_create (Expression **args, gint count, ContextType context)
{
  AskItem  *ask = g_new0 (AskItem, 1);

  set_param ("ask", args, count, 
	     (1<<EXP_VAR), &ask->name, 
	     0x10000 | (1<<EXP_STR), &ask->label,
	     0x10000 | (1<<EXP_STR), &ask->help,
	     0x20000 | (1<<EXP_VAR), &ask->type,
	     0x1FFFF, &ask->value, 
	     0);

  if (count > 5) {
    ask->constraints = g_new (Expression*, count - 5);
    memcpy (ask->constraints, &args[5], sizeof (Expression*) * count - 5);
  }
  ask->constraint_cnt = count - 5;

  ask->context = context;

  return ask;
}

/**********************************************************/
HdrItem*
hdritem_create (Expression **args, gint count, ContextType context)
{
  HdrItem  *hdr = g_new0 (HdrItem, 1);

  set_param ("hdr", args, count, 
	     0x1FFFF, &hdr->offset, 
	     0x1FFFF, &hdr->size, 
	     0x20000 | (1<<EXP_VAR), &hdr->type, 
	     0x1FFFF, &hdr->value, 
	     0);

  hdr->context = context;

  return hdr;
}

/**********************************************************/
SetItem*
setitem_create (Expression **args, gint count, ContextType context)
{
  SetItem  *set = g_new0 (SetItem, 1);

  set_param ("set", args, count, 
	     (1<<EXP_VAR), &set->name, 
	     0x1FFFF, &set->value, 
	     0x20000 | (1<<EXP_NUM) | (1<<EXP_VAR), &set->phase, 
	     0);

  if (!set->phase) set->phase = PHASE_DEFAULT;
  set->context = context;

  return set;
}


/**********************************************************/
#define GET_PARAM(n)   if (!strcasecmp (name, #n)) { \
                         set_param (#n, (Expression**)args, count, (1<<EXP_STR), &format->n, 0); \
}

/**********************************************************/
RawFormat*
rawformat_read (gchar *filename)
{
  FILE        *frm;
  gchar       *name;
  Expression  *args[100];
  gint        count;
  ContextType context;
  gchar       inbuf[MAX_INBUF_SIZE];
  RawFormat   *format;


  g_print ("format: %s\n", filename);

  frm = fopen (filename, "r");
  if (!frm) {
    WARNING ("Can't open file: '%s'\n", filename);
    return NULL;
  }

  inbuf[0]=0;
  fgets (inbuf, MAX_INBUF_SIZE, frm);
  if (strncasecmp (inbuf, RAW_FORMAT_MAGIC, strlen (RAW_FORMAT_MAGIC))) {
    WARNING ("Not a RAW format: '%s'\n", filename);
    return NULL;
  }

  format = rawformat_new ();
  format->file = g_strdup (filename);

  while (!feof (frm)) {
    if (!rawformat_get_parsed_line (frm, &name, (Expression**)args, &count, &context)) break;

    if (count < 0) break;

    GET_PARAM(name)
    else GET_PARAM(id)
    else GET_PARAM(desc)
    else GET_PARAM(author)
    else GET_PARAM(maintainer)
    else GET_PARAM(date)
    else GET_PARAM(loadmenu)
    else GET_PARAM(savemenu)
    else GET_PARAM(type)
    else GET_PARAM(header_fn)
    else GET_PARAM(pixel_fn)
    else GET_PARAM(drawable_fn)
    else GET_PARAM(image_fn)
    else if (!strcasecmp (name, "magic")) { 
      set_param (name, args, count, (1<<EXP_STR), &format->ext, (1<<EXP_STR), &format->magic, 0);
    }
  
    else if (!strcasecmp (name, "headersize")) { /* FIXME: EXP_VAR, EXP_LISP */
      set_param (name, args, count, 0x10000 | (1<<EXP_NUM), &format->headersize, 0);
    }
    else if (!strcasecmp (name, "ask")) {
      AskItem  *ask = askitem_create ((Expression**)args, count, context);
      g_list_append (format->ask_list, ask);
      if (context == CTX_SAVE || context == CTX_BOTH) format->ask_cnt_save++;
      if (context == CTX_LOAD || context == CTX_BOTH) format->ask_cnt_load++;
    }
    else if (!strcasecmp (name, "hdr")) {
      HdrItem  *hdr = hdritem_create ((Expression**)args, count, context);
      g_list_append (format->hdr_list, hdr);
    }
    else if (!strcasecmp (name, "set")) {
      SetItem  *set = setitem_create ((Expression**)args, count, context);
      g_list_append (format->set_list, set);
    }
  }
  fclose (frm);
  return format;
}

/**********************************************************/
void
rawformat_free (RawFormat *format)
{
  if (format) {
    g_free (format->name);
    g_free (format->id);
    g_free (format->desc);
    g_free (format->file);
    g_free (format->ext);
    g_free (format->author);
    g_free (format->maintainer);
    g_free (format->date);
    g_free (format->loadmenu);
    g_free (format->savemenu);
    g_free (format->magic);
    g_free (format->type);

    expression_free (format->headersize);

    /* FIXME: the list contents should be freed */
    if (format->hdr_list) g_list_free (format->hdr_list);
    if (format->ask_list) g_list_free (format->ask_list);
    if (format->set_list) g_list_free (format->set_list);

    g_free (format);
  }
}

/**********************************************************/
/* Code based on lsystem.c */
/* Code based on gflare.c */

/*
 *    Query gimprc for gflare-path, and parse it.
 *      This code is based on script_fu_find_scripts ()
 */

static GList*
parse_formats_path (void)
{
  GParam  *return_vals;
  gint    nreturn_vals;
  gchar   *path_string;
  gchar   *home;
  gchar   *path;
  gchar   *token;
  struct stat filestat;
  gint    err;
  GList   *path_list;
  gint    home_path_now=0;

  path_list = g_list_alloc ();

  return_vals = gimp_run_procedure ("gimp_gimprc_query",
				    &nreturn_vals,
				    PARAM_STRING, "raw-path",
				    PARAM_END);

  if (return_vals[0].data.d_status != STATUS_SUCCESS
      || return_vals[1].data.d_string == NULL) {
    gimp_destroy_params (return_vals, nreturn_vals);
    g_list_free (path_list);
    g_print ("Empty \"raw-path\" parameter in your gimprc\n");
    return NULL;
  }
  path_string = g_strdup (return_vals[1].data.d_string);
  gimp_destroy_params (return_vals, nreturn_vals);

  /* Set local path to contain temp_path, where (supposedly)
   * there may be working files.
   */
  home = getenv ("HOME");

  /* Search through all directories in the  path */

  token = strtok (path_string, ":");

  while (token) {
    home_path_now=0;
    if (*token == '\0') {
      token = strtok (NULL, ":");
      continue;
    }
    if (*token == '~') {
      path = g_malloc (strlen (home) + strlen (token) + 2);
      sprintf (path, "%s%s", home, token + 1);
    } else {
      path = g_malloc (strlen (token) + 2);
      strcpy (path, token);
    }				/* else */

    /* Check if directory exists */
    err = stat (path, &filestat);
    
    if (!err && ( (S_IWUSR & filestat.st_mode) != 0 ))
      home_path_now=1;

    if (!err && S_ISDIR (filestat.st_mode)) {
      if (path[strlen (path) - 1] != '/')
	strcat (path, "/");

      path_list = g_list_append (path_list, path);
      if (home_path_now && !raw_home_path) 
	raw_home_path=g_strdup(path);
    } else {
      g_free (path);
    }

    token = strtok (NULL, ":");
  }
  g_free (path_string);
  return path_list;
}

/**********************************************************/
static GList*
load_formats (GList *list)
{
  RawFormat  *format;
  GList      *frmlist;
  gchar      *path;
  gchar      *filename;
  DIR        *dir;
  struct dirent *dir_ent;
  struct stat   filestat;
  gint       err;
  gboolean   supply_slash;

  if (!list) return NULL;
  frmlist = g_list_alloc ();
  list = list->next;
  while (list) {
    path = list->data;
    list = list->next;
    supply_slash = (path[strlen (path) - 1] != '/');

    /* Open directory */
    dir = opendir (path);

    if (!dir)
      WARNING ("error reading RAW formats directory \"%s\"", path);
    else {
      while ((dir_ent = readdir (dir))) {
	filename = g_malloc (strlen (path) + strlen (dir_ent->d_name) + 1);
	if (supply_slash)
	  sprintf (filename, "%s/%s", path, dir_ent->d_name);
	else
	  sprintf (filename, "%s%s", path, dir_ent->d_name);


	/* Check for ".frm" extension */
	if (strcmp (&filename[strlen (filename) - 4], ".frm") != 0) {
	  g_free (filename);
	  continue;
	}
	/* Check the file and see that it is not a sub-directory */
	err = stat (filename, &filestat);

	if (!err && S_ISREG (filestat.st_mode)) {
	  format = rawformat_read (filename);
	  if (format)
	    g_list_append (frmlist, format);
	}
	g_free (filename);
      }				/* while */
      closedir (dir);
    }				/* else */
  }

  return frmlist;
}

/**********************************************************/
/* End of file format.c */
