/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/header.c,v 1.2 1999/01/31 22:32:26 benkovsk Exp benkovsk $
 */

/**********************************************************/
#include <stdio.h>
#include <string.h>

#include "RAW.h"

/**********************************************************/
static gboolean
header_save_basic (FILE *tgt, RawFormat *format);

static gboolean
header_load_basic (FILE *src, RawFormat *format);

/**********************************************************/
/* basic: save/load binary headers */

HeaderFnRec  header_fn_rec[] =
{
  /* name      save fn             load fn */
  { "basic",   header_save_basic,  header_load_basic }, 
};

gint  header_fn_rec_cnt = sizeof (header_fn_rec) / sizeof (HeaderFnRec);

/**********************************************************/
/**********************************************************/
void
header_find_function (ContextType context, gchar *header_name, 
		      HeaderFn *header_fn)
{
  gint      i;
  gboolean  Save;


  g_assert (context == CTX_SAVE || context == CTX_LOAD);

  Save = (context == CTX_SAVE);

  if (header_fn) {
    *header_fn = Save ? header_fn_rec[0].save_fn : header_fn_rec[0].load_fn;

    if (header_name) {
      for (i=0; i<header_fn_rec_cnt; i++)
	if (!strcmp (header_name, header_fn_rec[i].name)) {
	  *header_fn = Save ? header_fn_rec[i].save_fn : header_fn_rec[i].load_fn;
	  break;
	}
    }
  }
}

/**********************************************************/
static gboolean
header_save_basic (FILE *tgt, RawFormat *format)
{
  GList    *hlist;
  guchar   *header;
  VarNum   num, offset, size;
  gint     type, textlen;
  gchar    *text;
  gint     header_size;


  type = expression_evaluate (format->headersize, &header_size, &text, &textlen);
  if (type != EXP_NUM) return FALSE;
  if (!header_size) return TRUE;

  header = (guchar*) g_malloc (header_size);
  memset (header, 0, header_size);

  hlist = format->hdr_list->next;

  while (hlist) {
    HdrItem   *item = (HdrItem*) hlist->data;
    int       i;

    hlist = hlist->next;

    if (item->context == CTX_LOAD) continue;

    type = expression_evaluate (item->offset, &offset, &text, &textlen);
    type = expression_evaluate (item->size, &size, &text, &textlen);
    type = expression_evaluate (item->value, &num, &text, &textlen);

    switch (item->type) {
    case HDR_BIGEND:
      for (i=size; i; i--) {
	header[offset+i-1] = num % 0x100;
	num /= 0x100;
      }
      break;

    case HDR_LITTLEEND:
      for (i=0; i<size; i++) {
      	header[offset+i] = num % 0x100;
	num /= 0x100;
      }
      break;

    /* FIXME: another numeric type: strto* () */

    case HDR_STRING:
      memcpy (&header[offset], text, size?size:textlen);
      g_free (text);
      break;

    default:
      WARNING ("Unknown header item type: %d\n", item->type);
    }
  }

  if (header) {
    fwrite (header, header_size, 1, tgt);
    g_free (header);
  }

  return TRUE;
}

/**********************************************************/
static gboolean
header_load_basic (FILE *src, RawFormat *format)
{
  GList    *hlist;
  guchar   *header;
  VarNum   num, offset, size;
  gint     type, textlen;
  gchar    *text;
  gint     header_size;
  gboolean ok;


  type = expression_evaluate (format->headersize, &header_size, &text, &textlen);
  if (type != EXP_NUM) return FALSE;
  if (!header_size) return TRUE;

  header = (guchar*) g_malloc (header_size);
  fread (header, header_size, 1, src);

  hlist = format->hdr_list->next;

  while (hlist) {
    HdrItem   *item = (HdrItem*) hlist->data;
    int       i;

    hlist = hlist->next;

    if (item->context == CTX_SAVE) continue;

    type = expression_evaluate (item->offset, &offset, &text, &textlen);
    type = expression_evaluate (item->size, &size, &text, &textlen);

    num = 0;
    text = NULL;

    switch (item->type) {
    case HDR_BIGEND:
      for (i=0; i<size; i++) {
	num = 256*num + header[offset+i];
      }
      break;

    case HDR_LITTLEEND:
      for (i=size; i; i--) {
	num = 256*num + header[offset+i-1];
      }
      break;

    case HDR_STRING:
      text = (gchar*)g_malloc (size+1);
      memcpy (text, &header[offset], size);   /* size=0: nacist az do \0 */
      text[size]=0;                           /* for easy debugging */
      break;

    default:
      WARNING ("Unknown header item type: %d\n", item->type);
    }

    ok = expression_assign (item->value, num, text, size);
    if (!ok) {
      WARNING ("Expression assignment failed: %s\n", item->value->text);
    }

    if (text) g_free (text);
  }

  g_free (header);
  return TRUE;
}

/**********************************************************/
/* End of file header.c */
