/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * RAW plugin version 0.0.1
 * Copyright (C) 1999 Jaroslav Benkovsky, <benkovsk@pha.pvt.cz>
 *
 * RCS: $Header: /home/benkovsk/worx/RAW/RCS/variables.c,v 1.1 1999/01/31 22:31:15 benkovsk Exp benkovsk $
 */

/**********************************************************/
#include "RAW.h"

/**********************************************************/

#ifndef RAW_WITH_GUILE
GHashTable *variables = NULL;

/**********************************************************/
guint
hash_fn (gpointer key)
{
  /* FIXME: use gperf to get better fn */
  return  ((gchar*)key)[0] * strlen ((gchar*)key);
}

/**********************************************************/
gint
cmp_fn (gpointer a, gpointer b)
{
  return !strcmp ((gchar*)a, (gchar*)b);
}
#endif  /* ! RAW_WITH_GUILE */


/**********************************************************/
void
variables_init ()
{
#ifndef RAW_WITH_GUILE
  variables = g_hash_table_new (hash_fn, cmp_fn);
#endif
}

/**********************************************************/
void
variable_num_add (VarName *name, VarNum num)
{
#ifdef RAW_WITH_GUILE
  SCM   scm;

  scm = NUM2SCM (num);
  gh_define (name, scm);
#else
  Expression  *var = g_new0 (Expression, 1);

  var->type = EXP_NUM;
  var->num = num;

  g_hash_table_insert (variables, g_strdup (name), var);
#endif
}

/**********************************************************/
void
variable_float_add (VarName *name, VarFloat fnum)
{
#ifdef RAW_WITH_GUILE
  SCM   scm;

  scm = gh_double2scm (fnum);
  gh_define (name, scm);
#else
  Expression  *var = g_new0 (Expression, 1);

  var->type = EXP_FLOAT;
  var->fnum = fnum;

  g_hash_table_insert (variables, g_strdup (name), var);
#endif
}

/**********************************************************/
/* creates/assigns a variable with given string value of
 * specified size. If len < 0, strlen () is used instead.
 */
/**********************************************************/

void
variable_text_add (VarName *name, VarText *text, gint len)
{
#ifdef RAW_WITH_GUILE
  SCM   scm;

  scm = gh_str2scm (text, len>=0?len:strlen (text));
  gh_define (name, scm);
#else
  Expression  *var = g_new0 (Expression, 1);

  var->type = EXP_STR;
  if (len < 0) len = strlen (text);

  var->text = (gchar*) g_malloc (len + 1)
  memcpy (var->text, text, len);
  var->text[len] = '\0';
  var->len = len;

  g_hash_table_insert (variables, g_strdup (name), var);
#endif
}

/* FIXME: variable_float_add ? */

/**********************************************************/
void
variable_expr_add (VarName *name, Expression *expr)
{
  if (expr) {
    switch (expr->type) {
    case EXP_NUM: variable_num_add (name, expr->num); break;
    case EXP_STR: variable_text_add (name, expr->text, expr->len); break;
    case EXP_FLOAT: variable_float_add (name, expr->fnum); break;
    /* FIXME: EXP_LISP */
    default     : g_assert_not_reached ();
    }
  }
  else g_assert ("Expression == NULL\n");
}

/**********************************************************/
Expression*
variable_get (VarName *name)
{
  Expression  *value;

#ifdef RAW_WITH_GUILE
  SCM    scm;
  gchar  *buf;
  char   *req_template = "(false-if-exception %s)";


  buf = (gchar*)g_malloc (strlen (req_template) + strlen (name) - 1);
  sprintf (buf, req_template, name);

  scm = gh_eval_str (buf);
  g_free (buf);

  if (!scm) {
    WARNING ("variable_get (%s): SCM==NULL\n", name);
    return NULL;
  }

  if (scm == SCM_BOOL_F) {
    WARNING ("variable_get (%s): SCM==#f (Variable not defined)\n", name);
    return NULL;
  }

  value = g_new0 (Expression, 1);

  if (gh_exact_p (scm)) {
    value->num = SCM2NUM (scm);
    value->type = EXP_NUM;
  }
  else if (gh_inexact_p (scm)) {         /* FIXME: how to recognize float? */
    value->fnum = gh_scm2double (scm);
    value->type = EXP_FLOAT;
  }
  else {
    value->text = gh_scm2newstr (scm, &value->len);
    value->type = EXP_STR;
  }
  return value;
#else
  return expression_copy ((Expression*)g_hash_table_lookup (variables, 
							    (gpointer)name));
#endif
}

/**********************************************************/
gboolean
variable_get_num (VarName *name, VarNum *num)
{
  Expression  *expr = NULL;

  expr = variable_get (name);

  if (!expr) {
    WARNING ("Undefined variable: %s\n", name);
    goto error;
  }

  if (expr->type != EXP_NUM) {
    WARNING ("Variable is not numeric: %s\n", name);
    goto error;
  }

  *num = expr->num;
  expression_free (expr);
  return TRUE;

error:
  *num = 0;
  expression_free (expr);
  return FALSE;
}

/**********************************************************/
gboolean
variable_get_float (VarName *name, VarFloat *fnum)
{
  Expression  *expr = NULL;

  expr = variable_get (name);

  if (!expr) {
    WARNING ("Undefined variable: %s\n", name);
    goto error;
  }

  if (expr->type != EXP_FLOAT) {
    WARNING ("Variable is not float: %s\n", name);
    goto error;
  }

  *fnum = expr->fnum;
  expression_free (expr);
  return TRUE;

error:
  *fnum = 0;
  expression_free (expr);
  return FALSE;
}

/**********************************************************/
gboolean
variable_get_text (VarName *name, VarText **text, gint *len)
{
  Expression  *expr = NULL;

  expr = variable_get (name);

  if (!expr) {
    WARNING ("Undefined variable: %s\n", name);
    goto error;
  }

  if (expr->type != EXP_STR) {
    WARNING ("Variable is not string: %s\n", name);
    goto error;
  }

  *text = expr->text;
  expr->text = NULL;
  *len = expr->len;
  expression_free (expr);
  return TRUE;

error:
  *text = NULL;
  *len = 0;
  expression_free (expr);
  return FALSE;
}

/**********************************************************/
/* End of file variables.c */
